'use strict';
/**
 * node-windows-window-controller
 *
 * @author waiting
 * @license MIT
 * @link https://github.com/waitingsong/node-windows-window-controller
 */
Object.defineProperty(exports, '__esModule', { value: true });
/// <reference types="node" />
const ffi = require('ffi');
// import * as ref from 'ref';
const isWin32 = process.platform === 'win32';
const plateformError = 'Invalid platform: win32 required';
// https://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx
var CmdShow;
(function (CmdShow) {
    // Hides the window and activates another window.
    CmdShow[CmdShow['SW_HIDE'] = 0] = 'SW_HIDE';
    // Activates and displays a window. If the window is minimized or maximized, the system restores it to its original size and position. An application should specify this flag when displaying the window for the first time.
    CmdShow[CmdShow['SW_SHOWNORMAL'] = 1] = 'SW_SHOWNORMAL';
    // Activates the window and displays it as a minimized window.
    CmdShow[CmdShow['SW_SHOWMINIMIZED'] = 2] = 'SW_SHOWMINIMIZED';
    // Activates the window and displays it as a maximized window.
    CmdShow[CmdShow['SW_SHOWMAXIMIZED'] = 3] = 'SW_SHOWMAXIMIZED';
    // Maximizes the specified window.
    CmdShow[CmdShow['SW_MAXIMIZE'] = 3] = 'SW_MAXIMIZE';
    // Displays a window in its most recent size and position. This value is similar to SW_SHOWNORMAL, except that the window is not activated.
    CmdShow[CmdShow['SW_SHOWNOACTIVATE'] = 4] = 'SW_SHOWNOACTIVATE';
    // Activates the window and displays it in its current size and position.
    CmdShow[CmdShow['SW_SHOW'] = 5] = 'SW_SHOW';
    // Minimizes the specified window and activates the next top-level window in the Z order.
    CmdShow[CmdShow['SW_MINIMIZE'] = 6] = 'SW_MINIMIZE';
    // Displays the window as a minimized window. This value is similar to SW_SHOWMINIMIZED, except the window is not activated.
    CmdShow[CmdShow['SW_SHOWMINNOACTIVE'] = 7] = 'SW_SHOWMINNOACTIVE';
    // Displays the window in its current size and position. This value is similar to SW_SHOW, except that the window is not activated.
    CmdShow[CmdShow['SW_SHOWNA'] = 8] = 'SW_SHOWNA';
    // Activates and displays the window. If the window is minimized or maximized, the system restores it to its original size and position.
    CmdShow[CmdShow['SW_RESTORE'] = 9] = 'SW_RESTORE';
    // Sets the show state based on the SW_ value specified in the STARTUPINFO structure passed to the CreateProcess function by the program that started the application.
    CmdShow[CmdShow['SW_SHOWDEFAULT'] = 10] = 'SW_SHOWDEFAULT';
    // Minimizes a window, even if the thread that owns the window is not responding. This flag should only be used when minimizing windows from a different thread.
    CmdShow[CmdShow['SW_FORCEMINIMIZE'] = 11] = 'SW_FORCEMINIMIZE';
})(CmdShow = exports.CmdShow || (exports.CmdShow = {}));
const api = ffi.Library('user32.dll', {
    EnumWindows: ['bool', ['pointer', 'int32']],
    GetWindowThreadProcessId: ['uint32', ['uint', 'pointer']],
    GetWindowTextW: ['long', ['long', 'CString', 'long']],
    IsWindowVisible: ['bool', ['int32']],
    ShowWindow: ['bool', ['uint32', 'int']],
    GetParent: ['uint32', ['uint32']],
    GetAncestor: ['uint', ['uint32', 'uint']],
});
const config = {
    tno: 0,
    task: new Map(),
};
// stop loop if return false
const enumWindowsProc = ffi.Callback('bool', ['uint32', 'int'], (hWnd, lParam) => {
    const task = config.task.get(lParam);
    if (!task) {
        return true;
    }
    if (task.pid > 0) {
        const buf = Buffer.alloc(8);
        api.GetWindowThreadProcessId(hWnd, buf);
        const pid = buf.readUIntLE(0, 8);
        if (pid && pid === task.pid) {
            task.hwndSet.add(hWnd);
        }
    }
    else {
        if (task.title) {
            const buf = Buffer.alloc(254);
            const ret = api.GetWindowTextW(hWnd, buf, 254);
            const name = buf.toString('ucs2');
            // const visible = api.IsWindowVisible(hWnd);
            if (name.indexOf(task.title) !== -1) {
                const buf = Buffer.alloc(8);
                api.GetWindowThreadProcessId(hWnd, buf);
                task.pidSet.add(buf.readUIntLE(0, 8));
                task.hwndSet.add(hWnd);
                // console.log(`${hWnd}: ${name}`);
            }
        }
        else {
            task.hwndSet.add(hWnd);
        }
    }
    return true;
});
// expose original user32.ShowWindow() as default
function showWindow(hWnd, nCmdShow) {
    let errcode = 1;
    if (!Number.isSafeInteger(hWnd)) {
        console.error('hWnd must integer');
        return errcode;
    }
    if (!validate_cmdshow(nCmdShow)) {
        return errcode;
    }
    try {
        if (api.ShowWindow(hWnd, nCmdShow)) {
            errcode = 0;
        }
    }
    catch (ex) {
        console.error(ex);
    }
    return errcode;
}
exports.default = showWindow;
function validate_cmdshow(nCmdShow) {
    const res = (nCmdShow < 0 || typeof CmdShow[nCmdShow] === 'undefined') ? false : true;
    if (!res) {
        console.error('hWnd value invalid: See: https://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx');
    }
    return res;
}
exports.validate_cmdshow = validate_cmdshow;
function hide(p, onlyMainWin = true) {
    return proxy(p, CmdShow.SW_HIDE, onlyMainWin);
}
exports.hide = hide;
function show(p, nCmdShow, onlyMainWin = true) {
    return proxy(p, nCmdShow, onlyMainWin);
}
exports.show = show;
// https://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx
function proxy(p, nCmdShow, onlyMainWin) {
    const execRet = init_execret();
    if (!validate_cmdshow(nCmdShow)) {
        execRet.err = 1;
        execRet.msg = 'value of nCmdShow invalid';
        return Promise.resolve(execRet);
    }
    return new Promise(resolve => {
        if (onlyMainWin) {
            get_main_hwnd(p).then(hWnd => {
                if (hWnd && typeof hWnd === 'number') {
                    _show(hWnd, nCmdShow, onlyMainWin)
                        .then((hWnd) => {
                        hWnd && execRet.hwnds.push(hWnd);
                    })
                        .catch(err => {
                        execRet.err = 1;
                        execRet.msg += '\n ' + err;
                    })
                        .then(() => {
                        resolve();
                    });
                }
                else {
                    resolve();
                }
            });
        }
        else {
            get_hwnds(p).then(arr => {
                if (arr && arr.length) {
                    Promise.all(arr.map(hWnd => {
                        return _show(hWnd, nCmdShow, onlyMainWin)
                            .then((hWnd) => hWnd && execRet.hwnds.push(hWnd))
                            .catch(err => {
                            execRet.err = 1;
                            execRet.msg += '\n ' + err;
                        });
                    }))
                        .then(resolve);
                }
                else {
                    resolve();
                }
            });
        }
    }).then(() => execRet);
}
function _show(hWnd, nCmdShow, onlyMainWin) {
    return new Promise((resolve, reject) => {
        const id = +hWnd;
        if (onlyMainWin && !is_main_window(hWnd)) {
            return resolve();
        }
        nCmdShow = +nCmdShow;
        if (!id || Number.isNaN(id)) {
            return reject('_show() params hWnd invalid: ' + hWnd);
        }
        if (Number.isNaN(nCmdShow) || nCmdShow < 0) {
            return reject('_show() params nCmdShow invalid: ' + nCmdShow);
        }
        if (nCmdShow === 0) {
            if (!api.IsWindowVisible(id)) {
                return resolve();
            }
        }
        try {
            api.ShowWindow(id, nCmdShow);
        }
        catch (ex) {
            return reject(ex);
        }
        resolve(id);
    });
}
function get_main_hwnd(p) {
    return get_hwnds(p).then(arr => {
        if (arr && Array.isArray(arr) && arr.length) {
            return filter_main_hwnd(arr);
        }
    });
}
exports.get_main_hwnd = get_main_hwnd;
function get_hwnds(p) {
    return _get_hwnds(p);
}
exports.get_hwnds = get_hwnds;
function _get_hwnds(p, task) {
    if (!task) {
        task = create_task();
    }
    let t;
    return Promise.race([
        new Promise(resolve => {
            t = setTimeout(() => {
                console.error('timeout failed');
                resolve();
            }, 30000); // @HARDCOD
        }),
        _get_hwnd(p, task),
    ]).then(res => {
        clearTimeout(t);
        task && config.task.delete(task.tno);
        if (res && Array.isArray(res) && res.length) {
            return res;
        }
    }).catch(err => {
        clearTimeout(t);
        task && config.task.delete(task.tno);
        console.error(err);
    });
}
function _get_hwnd(p, task) {
    if (typeof p === 'number') {
        if (!Number.isSafeInteger(p)) {
            return Promise.resolve();
        }
        task.pid = p;
        task.title = '';
        return get_task_hwnd(task)
            .then(res => {
            return res;
        })
            .catch(err => {
            console.error(err);
            return;
        });
    }
    else if (typeof p === 'string') {
        task.pid = 0;
        task.title = p;
        return get_task_hwnd(task)
            .then(res => {
            return res;
        })
            .catch(err => {
            console.error(err);
            return;
        });
    }
    else {
        console.error('param p invalid', p);
        return Promise.resolve();
    }
}
function get_task_hwnd(task) {
    return new Promise((resolve, reject) => {
        if (!isWin32) {
            reject(plateformError);
            return;
        }
        if (!task) {
            return resolve([]);
        }
        api.EnumWindows.async(enumWindowsProc, task.tno, err => {
            resolve(Array.from(task.hwndSet) || []);
        });
    });
}
// get hWnd of main top-level window
function filter_main_hwnd(arr) {
    const res = new Set();
    const map = new Map();
    for (let hWnd of arr) {
        const ownerHWnd = api.GetAncestor(hWnd, 3);
        let count = map.get(ownerHWnd);
        count = !count ? 1 : count + 1;
        map.set(ownerHWnd, count);
    }
    if (map.size === 1) {
        const [hWnd] = map.keys();
        return hWnd;
    }
    else if (map.size > 1) {
        let max = 0;
        let hWnd = 0;
        for (let [k, v] of map) {
            if (v > max) {
                max = v;
                hWnd = k;
            }
        }
        return hWnd;
    }
}
// check hWnd is top-level window
function is_main_window(hWnd) {
    const res = api.GetParent(hWnd);
    return res === null || res === 0 ? true : false;
}
// kill process by which the matched hWnd(s) (window) created
function kill(p) {
    const execRet = init_execret();
    if (typeof p === 'number') {
        _kill(p);
        execRet.pids.push(p);
        return Promise.resolve(execRet);
    }
    else if (typeof p === 'string') {
        return new Promise(resolve => {
            const task = create_task();
            return _get_hwnds(p, task).then(() => {
                if (task.pidSet.size) {
                    for (let pid of task.pidSet) {
                        if (_kill(pid)) {
                            execRet.pids.push(pid);
                        }
                    }
                }
                else {
                    execRet.msg = 'the pid list to be killed empty';
                }
                resolve(execRet);
            });
        });
    }
    else {
        execRet.err = 1;
        execRet.msg = 'kill() param must be either number or string';
        return Promise.resolve(execRet);
    }
}
exports.kill = kill;
function _kill(pid) {
    try {
        process.kill(pid, 0) && process.kill(pid);
        return true;
    }
    catch (ex) {
        console.error(ex);
    }
    return false;
}
function create_task() {
    const tno = ++config.tno;
    const task = {
        tno,
        pid: 0,
        title: '',
        hwndSet: new Set(),
        pidSet: new Set(),
    };
    config.task.set(tno, task);
    return task;
}
function init_execret() {
    return {
        err: 0,
        msg: '',
        pids: [],
        hwnds: [],
    };
}
